//
//    Copyright 2011 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "DominoDeviceManager.h"

#include <QHostAddress>
#include <QStringList>
#include <QMap>
#include <QList>
#include <QMetaType>
#include <QDebug>

#include "dominodevice.h"
#include "DeviceProperties.h"
#include "DeviceEntry.h"
#include "EventData.h"
#include "util/QDateTimeCompat.h"

const int monitorEventTimeMs = 300;

Q_DECLARE_METATYPE(QList<QHostAddress>)

DominoDeviceManager::DominoDeviceManager(QObject *parent) :
    QObject(parent),
    eventData(new EventData),
    monitorEvent(new EventData)
//    properties(new DeviceProperties())
{
    qRegisterMetaType<QList<QHostAddress> >();
    qRegisterMetaType<EventData>();
}

DominoDeviceManager::~DominoDeviceManager()
{
    devices.clear();
}

void DominoDeviceManager::setHostAddressList(const QList<QHostAddress> &list)
{
    devices.clear();
    Q_FOREACH(QHostAddress hostAddress, list)
    {
        DominoDevice * const dev = new DominoDevice();
        QString key = hostAddress.toString();
        DeviceEntry ent;
        ent.dev = DominoDevicePtr(dev);
        devices.insert(key, ent);
        connect(this, SIGNAL(setupChanged(DominoSettings)), dev, SLOT(on_setupChanged(DominoSettings)));
        connect(this, SIGNAL(clearEventNumber()), dev, SLOT(clearEventNumber()));
        connect(dev, SIGNAL(deviceConnected()), this, SLOT(onDeviceConnect()));
        connect(dev, SIGNAL(deviceDisconnected()), this, SLOT(onDeviceDisconnect()));
        connect(dev, SIGNAL(TempChanged(double)), SLOT(onDeviceTempChanged(double)));
        connect(dev, SIGNAL(numberOfChannelChanged(int)), SLOT(onDeviceNumberOfChannelChanged(int)));
        connect(dev, SIGNAL(adcCalibrationStatusChanged()), SLOT(onDeviceAdcCalibrationStatusChanged()));
        connect(dev, SIGNAL(startAdcCalibration()), SLOT(onDeviceAdcCalibrationStart()));
        connect(dev, SIGNAL(dataAcquired(DominoData)), SLOT(onDeviceDataAcquired(DominoData)));
        QMetaObject::invokeMethod(dev, "setHostAddress", Q_ARG(QHostAddress, hostAddress));
    }
//    qDebug() << "DominoDeviceManager::setHostAddressList:" << devices.size() << "devices created";
}

//void DominoDeviceManager::clearEventNumber()
//{
//    Q_FOREACH(DeviceEntry ent, devices)
//        QMetaObject::invokeMethod(ent.dev.data(), "clearEventNumber");
//}

int DominoDeviceManager::getChannelMaxSampleCount() const
{
    if (devices.isEmpty()) return 0;
    const DeviceEntry &first = *devices.begin();
    const DominoDevicePtr &firstdev = first.dev;
    int result = firstdev->getChannelMaxSampleCount();
    Q_FOREACH(DeviceEntry ent, devices) {
        int n = ent.dev->getChannelMaxSampleCount();
        if (n < result)
            result = n;
    }
    return result;
}

bool DominoDeviceManager::isConnected() const
{
    if (devices.isEmpty())
        return false;
    Q_FOREACH(DeviceEntry ent, devices)
        if (!ent.dev->isConnected())
            return false;
    return true;
}

//QtMregDevice::DeviceId DominoDeviceManager::getDeviceId() const
//{
//}

QString DominoDeviceManager::getDeviceIdStr() const
{
    QStringList sl;
    Q_FOREACH(DeviceEntry ent, devices)
        sl.push_back(ent.dev->getDeviceIdStr());
    return sl.join("\n");
}

void DominoDeviceManager::onDeviceConnect()
{
    if (isConnected()) {
        qDebug() << "DominoDeviceManager: all devices connected";
        emit deviceConnected();
    }
}

void DominoDeviceManager::onDeviceDisconnect()
{
    if (!isConnected()) {
        qDebug() << "DominoDeviceManager: any device disconnected";
        emit deviceDisconnected();
    }
}

void DominoDeviceManager::clearFullEventNumber()
{
    eventData->clearFullEventNumber();
}

void DominoDeviceManager::onDeviceTempChanged(double value)
{
    QStringList sl;
//    qDebug() << "TempChanged: " << value;
    Q_FOREACH(QString key, devices.keys()) {
        const DominoDevicePtr &dev = devices.value(key).dev;
        double temp = dev->getTemperature();
//        devices.value(key).prop.temperature;
//        const DominoDevicePtr &dev = devices.value(key).dev;
//        properties[key].temperature = value;
        sl.push_back(QString("%1 %2C").arg(temp, 0, 'f', 1).arg(QChar(0x02da)));
    }
    QString str = sl.join("\n");
//    qDebug() << "TempChanged: " << str;
    emit tempChanged(str);
}

void DominoDeviceManager::onDeviceNumberOfChannelChanged(int value)
{
//    qDebug() << "DominoDeviceManager::onDeviceNumberOfChannelChanged: devices.size = " << devices.size();
    int n = 0;
    Q_FOREACH(DeviceEntry ent, devices) {
        const DominoDevicePtr &dev = ent.dev;
        n += dev->getNumberOfChannels();
//        qDebug() << QString("DominoDeviceManager::onDeviceNumberOfChannelChanged: n %1").arg(dev->getNumberOfChannels());
    }
    emit numberOfChannelChanged(n);
}


void DominoDeviceManager::onDeviceAdcCalibrationStatusChanged()
{
    bool status_all = true;
    Q_FOREACH(DeviceEntry ent, devices) {
        const DominoDevicePtr &dev = ent.dev;
        status_all &= dev->getAdcCalibrationStatus();
    }
    emit adcCalibrationStatusChanged(status_all);
}

void DominoDeviceManager::onDeviceAdcCalibrationStart()
{
    emit adcCalibrationStart();
}

void DominoDeviceManager::maybeSendMonitorEvent(const EventData &eventData)
{
    qint64 dt = QDateTimeMsescTo(monitorEvent->timestamp, eventData.timestamp);
//    qDebug() << "monitorEvent dt = " << dt;
    if (dt > monitorEventTimeMs) {
        *monitorEvent = eventData;
        emit monitorEventChanged(*monitorEvent);
//    } else {
//        qDebug() << "monitorEvent suppressed";
    }
}

void DominoDeviceManager::onDeviceDataAcquired(const DominoData &dominoData)
{
    quint64 key = dominoData.serial_id;

//    if (eventData->find(key) != eventData->end()) {
//    } else {
//    }
    if (eventData->size() > devices.size()) {
        qWarning() << "Event demultiplexer: Too much different id numbers, event buffer reset";
        eventData->clear();
    }
    (*eventData)[key] = dominoData;

    bool match = eventData->timestampTrigMatch();
    // emit data only when all event numbers match
    if (!match) {
        qWarning() << "Timestamp mismatch";
        return;
    }
    if (eventData->size() == devices.size()) {
        eventData->timestamp = QDateTime::currentDateTime();
        eventData->FullEventNumber++;
        maybeSendMonitorEvent(*eventData);
        emit dataAcquired(*eventData);
        eventData->clear();
        //        qDebug() << "Event number Ok";
    } else {
        qWarning() << QString("[TS:%1, ID:%2]: Number of event blocks mismatch: %3 blocks, %4 devices")
                      .arg(dominoData.trig_ts)
                      .arg(dominoData.serial_id, 4, 16, QChar('0'))
                      .arg(eventData->size()).arg(devices.size());
    }
}
