
#include "regio/mlinkdevice.h"
#include "regio/mregdevice.h"
#include "dominodevice.h"
#include <stdexcept>
#include "dominodata.h"
#include <QDebug>
#include <QString>
#include <QVector>
#include <QMessageBox>
#include "util/QDateTimeCompat.h"

//const int ADC_CHANNELS = 64;
//const int MAX_MEMRD_SIZE = 100;       //in 32bit words
const int ADC_BITS = 12;
const int ADC_HALF_VALUE = 1<<(ADC_BITS-1);
const int ADC_MASK = (1<<ADC_BITS) - 1;
const int MASTER_SLAVE_TRIG_DELAY = 35;

const int pollTimeMs = 10;

enum { BOARD_REG_BITS = 13 };
const quint16 BOARD_REG_COUNT = (1<<BOARD_REG_BITS);

enum { MEM_BIT_SELECT_CTRL = 1<<13 };

// Control registers
enum {
    ADCM16_CTRL_REG           = 0x40,
    ADCM16_WIN_REG            = 0x41,
    ADCM16_STATUS_REG         = 0x42, /* read-only */
    ADCM16_TRIGMULT_REG       = 0x43,
    ADCM16_FRAGMLEN_REG       = 0x44,
    ADCM16_FRAGMMAX_REG       = 0x45, /* read-only */
    ADCM16_FIFO_PFULL_REG     = 0x46,
    ADCM16_MAX_NFRAG_REG      = 0x47, /* read-only */
    ADCM16_EXT_LAT_REG  = ADCM16_WIN_REG,
    ADCM16_EXT_MATCHWIN_REG   = 0x48,
    ADCM16_EXT_PREWIN_REG     = 0x49,
    ADCM16_CH_DPM_KS_REG      = 0x4A, /* read-only */
    ADCM16_TEMPERATURE_REG    = 0x4B, /* read-only */
    ADCM16_FW_VER_REG         = 0x4C, /* read-only */
    ADCM16_FW_REV_REG         = 0x4D, /* read-only */
    ADCM16_SERIAL_ID_REG      = 0x4E, /* read-only */
    ADCM16_PROPERTIES_REG     = 0x4F,  /* read-only */

    // Timestamp registers
    ADCM16_TS_SET_REG64       = 0x54,
    ADCM16_TS_READ_REG64      = 0x5C
};

enum {
    REG_RUN_STATUS = ADCM16_STATUS_REG,//0x00,
    REG_DEVICE_CTRL = ADCM16_CTRL_REG,//0x40, //64
    REG_DEVICE_RLAT = ADCM16_WIN_REG, //0x41, //65
    REG_DEVICE_ID = ADCM16_STATUS_REG, //0x42, // 66
    REG_SERIAL_NUM = ADCM16_SERIAL_ID_REG, //0x50, // 80
    REG_FW_VER = ADCM16_FW_VER_REG, //0x56, // 86
    REG_FW_REV = ADCM16_FW_REV_REG, //0x57, // 87
    REG_TEMPERATURE = ADCM16_TEMPERATURE_REG, //0x54, // 84
    REG_TRIG_CTRL = ADCM16_TRIGMULT_REG,//0x6E, // 110
    REG_DAC_MAX550_1 = 0x100, // 256
    REG_DAC_MAX550_2 = 0x101, // 257
    REG_PCA_12 = 0x102, // 258
    REG_ADC_SPI = 0x103, //259
    REG_ADC12_SPI_READ = 0x104, //260
    REG_ADC34_SPI_READ = 0x105, //261
    REG_ADC56_SPI_READ = 0x106, //262
    REG_ADC78_SPI_READ = 0x107, //263
    REG_ADC_CLK_CTRL = 0x108, //264
    REG_ZS_EVENTS    = 0x110,  //(64 bits) 272,273,274,275

    // 0x120-0x123 - AD9252 SPI device #0
    REG_AD9252_SPI_ADDR = 0x120, // 288
    REG_AD9252_SPI_WR_DATA = 0x121,
    REG_AD9252_SPI_READ_DATA = 0x122,
    REG_AD9252_CSR = 0x123,
    // 0x123 reserved
    // 0x124-0x127 - AD9252 SPI device #1

    REG_AMP_SET = 0x130,
    REG_AD5622_COMMAND = 0x131,
    REG_AD5622_BASELINE = 0x132,
    REG_AD5622_LEVEL = 0x133,

    REG_GTPDUAL = 0x200 //512
};

enum {
    MEM_CH_WR_ADDR  = 0x0000,
    MEM_CH_CTRL     = 0x0001,
    MEM_CH_THR      = 0x0002,
    MEM_CH_ZS_THR   = 0x0003,
    MEM_CH_ADC_DATA = 0x0005,
    MEM_CH_ADC_PATTERN = 0x0006,
    MEM_CH_ADC_PATTERN_MISMATCH_CNT = 0x0007
};

enum {
    REG_DEVICE_CTRL_BIT_RUN = 0x0001,
    REG_DEVICE_CTRL_BIT_INC_BITCLK_DELAY = 0x0008
};

enum {
    REG_RUN_STATUS_BIT_SLAVE_0 = 0x0001,
    REG_RUN_STATUS_BIT_SLAVE_1 = 0x0002,
    REG_RUN_STATUS_BIT_SLAVE_2 = 0x0004,
    REG_RUN_STATUS_BIT_SLAVE_3 = 0x0008,
    REG_RUN_STATUS_BIT_RUNNING = 0x0010 //0x0001
};

enum {
    MEM_CH_CTRL_BIT_EN                = 0x0001,
    MEM_CH_CTRL_BIT_INVERT            = 0x0002,
    MEM_CH_CTRL_BIT_THR_TRIG_INVERT   = 0x0004,
    MEM_CH_CTRL_BIT_ZS_THR_INVERT     = 0x0008
};

enum {
    ADC_CLK_CTRL_BIT_INC   = 0x8000,
    ADC_CLK_CTRL_BIT_RESET = 0x0100,
    ADC_CLK_CTRL_BIT_ADC_RESET = 0x4000,
    ADC_CLK_CTRL_ALL_ADCS  = 0x00FF
};

enum {
    ADC_SPI_PATTERN_LOW_REG   = 0x10,
    ADC_SPI_PATTERN_HIGH_REG  = 0x12,
    ADC_SPI_DECIMATOR_CTRL_REG = 0x16
};

// AD9252
enum {AD9252_CHIP_ID = 0x09};
enum {
    AD9252_SPI_12BITS_MODE = 0x3
};

enum {
    AD9252_SPI_DEVICE_UPDATE_BIT = 0x1
};

enum {
    AD9252_SPI_RESET = 0x3C,
    AD9252_SPI_RUN = 0x18
};


enum {
    AD9252_SPI_CHIP_PORT_CONFIG_REG = 0x0,
    AD9252_SPI_CHIP_ID_REG = 0x1,
    AD9252_SPI_CHIP_GRADE_REG = 0x2,
    AD9252_SPI_SERIAL_CONTROL_REG = 0x21,
    AD9252_SPI_DEVICE_UPDATE_REG = 0xFF
};

enum {AD5622_WRITE_COMMAND = 0x180};

quint32 chBaseMemAddr(quint32 ch)
{
    return ch << 14;
}

DominoDevice::DominoDevice(QObject *parent)
    : QtMregDevice(parent),
      pollTimer(new QTimer(this)),
      setup(DominoSettings())
{
    addSupportedDeviceId(QtMregDevice::DEVICE_ID_ADC64);
    addSupportedDeviceId(QtMregDevice::DEVICE_ID_ADC8E);
    addSupportedDeviceId(QtMregDevice::DEVICE_ID_ADCM163);
    connect(this, SIGNAL(deviceConnected()), SLOT(on_dominoDeviceConnected()));
//    connect(this, SIGNAL(startAdcCalibration()), SLOT(on_startAdcCalibration()));
    connect(pollTimer, SIGNAL(timeout()), SLOT(on_timerPoll()));
    pollTimer->setSingleShot(true);
    pollTimer->start(pollTimeMs);
}

DominoDevice::~DominoDevice()
{
    disable();
}

int DominoDevice::getNumberOfChannels() const
{
    int nch = 1;
    switch (getDeviceId()) {
    case QtMregDevice::DEVICE_ID_TEST: break;
    case QtMregDevice::DEVICE_ID_ADC64: nch =  64; break;
    case QtMregDevice::DEVICE_ID_ADC8E: nch = 8; break;
    case QtMregDevice::DEVICE_ID_8CADC: nch = 8; break;
    case QtMregDevice::DEVICE_ID_ADCM16: nch = 16; break;
    case QtMregDevice::DEVICE_ID_ADC16LTCM: nch = 16; break;
    case QtMregDevice::DEVICE_ID_ADC16LTCS: nch = 16; break;
    case QtMregDevice::DEVICE_ID_ADCM163: nch = 16; break;
    case QtMregDevice::DEVICE_ID_BPM4100: nch = 4; break;
    case QtMregDevice::DEVICE_ID_EVADC: nch = 8; break;
    case QtMregDevice::DEVICE_ID_NUCLORF: break;
    case QtMregDevice::DEVICE_ID_BRIC: break;
    case QtMregDevice::DEVICE_ID_AnyDevice: break;
    }
    //    qDebug() << "nch = " << nch;
    return nch;
}

int DominoDevice::getChannelMaxSampleCount() const
{
    switch (getDeviceId()) {
    case QtMregDevice::DEVICE_ID_ADC64:   return 0x0800; // 2k
    case QtMregDevice::DEVICE_ID_ADC8E:   return 0x2000; // 8k
    case QtMregDevice::DEVICE_ID_ADCM163: return 0x1000; // 4k
    }
    return 0;
}

void DominoDevice::on_dominoDeviceConnected()
{
    settingsCacheValid = false;
    DetectNumberOfAdc64boards();
    testChannelRegisters();
    emit numberOfChannelChanged(getNumberOfChannels());
    emit startAdcCalibration(); // why signal?
    setup_adc(0);
    setup_adc(1);
    adcCalibrationStatus = FindAdcClkDelay();
    emit adcCalibrationStatusChanged();
}

void DominoDevice::ad9252_spi_write(quint16 addr, quint8 data, int device)
{
    regWrite(device*4 + REG_AD9252_SPI_WR_DATA, data);
    regWrite(device*4 + REG_AD9252_SPI_ADDR, addr & 0x1FFF);   //bits 13,14,15 = 0 (write and W0=W1=0 - 1 byte operation)
}

quint16 DominoDevice::ad9252_spi_read(quint16 addr, int device)
{
    regWrite(device*4 + REG_AD9252_SPI_ADDR, ((addr&0x9FFF) | 0x8000));   //bits 13,14 = 0 , bit15 = 1 (read and W0=W1=0 - 1 byte operation)
    return regRead(device*4 + REG_AD9252_SPI_READ_DATA);
}

void DominoDevice::setup_adc(int device)
{
    if (getDeviceId() != QtMregDevice::DEVICE_ID_ADCM163) return;
    // read ID
      quint8 chip_id = ad9252_spi_read(AD9252_SPI_CHIP_ID_REG, device);
      if (chip_id != AD9252_CHIP_ID) {
          qCritical() << QString("AD9252 SPI [%1]: failed to read ID: expected %2, got %3")
                         .arg(device)
                         .arg(AD9252_CHIP_ID)
                         .arg(chip_id);
          return;
      }
      quint8 chip_grade = ad9252_spi_read(AD9252_SPI_CHIP_GRADE_REG, device);
      qDebug() << QString("AD9252 SPI [%1]: chip id %2h, grade %3h")
                     .arg(device)
                  .arg(chip_id, 2, 16, QChar('0'))
                  .arg(chip_grade, 2, 16, QChar('0'));
      // Reset
      ad9252_spi_write(AD9252_SPI_CHIP_PORT_CONFIG_REG, AD9252_SPI_RESET, device);
      const int reset_watchdog = 10;
      int watchdog = 0;
      while(true)
      {
          if (ad9252_spi_read(AD9252_SPI_CHIP_PORT_CONFIG_REG, device) == AD9252_SPI_RUN) break;
          watchdog++;
          if(watchdog > reset_watchdog)
          {
              qCritical() << QString("AD9252 SPI [%1]: reset failed").arg(device);
              return;
          }
      }
      qDebug() << QString("AD9252 SPI [%1]: reset").arg(device);

/*
      ad9252_spi_write(0xD,  0x48, device);
      ad9252_spi_write(0x19, 0x21, device);
      ad9252_spi_write(0x1A, 0xF3, device);
      ad9252_spi_write(0x1B, 0x54, device);
      ad9252_spi_write(0x1C, 0xF6, device);
      ad9252_spi_write(AD9252_SPI_DEVICE_UPDATE_REG, AD9252_SPI_DEVICE_UPDATE_BIT, device);
*/

      quint8 serial_ctrl = ad9252_spi_read(AD9252_SPI_SERIAL_CONTROL_REG, device);
      if (serial_ctrl != AD9252_SPI_12BITS_MODE) {
          ad9252_spi_write(AD9252_SPI_SERIAL_CONTROL_REG, AD9252_SPI_12BITS_MODE, device);
          ad9252_spi_write(AD9252_SPI_DEVICE_UPDATE_REG, AD9252_SPI_DEVICE_UPDATE_BIT, device);
          if (ad9252_spi_read(AD9252_SPI_SERIAL_CONTROL_REG, device) != AD9252_SPI_12BITS_MODE) {
              qCritical() << QString("AD9252 SPI [%1]: failed to set serial mode").arg(device);
              return;
          } else {
              qDebug() << QString("AD9252 SPI [%1]: set serial mode").arg(device);
          }
      }
      // set offset binary (0) or two's complement output mode (1)
//      ad9252_spi_write(0x14, 0x01, device);
//      ad9252_spi_write(AD9252_SPI_DEVICE_UPDATE_REG, AD9252_SPI_DEVICE_UPDATE_BIT, device);

      // reset deserializer
      regWrite(device*4 + REG_AD9252_CSR, 1);
      regWrite(device*4 + REG_AD9252_CSR, 0);
      quint8 adc_status = regRead(device*4 + REG_AD9252_CSR) >> 8;
      if (adc_status != 0xa1) {
          qCritical() << QString("AD9252 SPI [%1]: serdes unlocked (%2)").arg(device).arg(adc_status, 2, 16, QChar('0'));
          return;
      } else {
          qDebug() << QString("AD9252 SPI [%1]: serdes locked").arg(device);
      }
      qDebug() << QString("AD9252 SPI [%1]: setup Ok").arg(device);
      quint16 tmp = regRead(REG_DEVICE_ID);
      qDebug() << QString("status: %1").arg(tmp, 2, 16, QChar('0'));
}


void DominoDevice::on_startAdcCalibration()
{

}

void DominoDevice::DetectNumberOfAdc64boards()
{
    if (getDeviceId() == QtMregDevice::DEVICE_ID_ADC64) {
        const quint16 gptStatus = regRead(REG_GTPDUAL);
            qDebug() << QString("GTP link status %1").arg(gptStatus, 6, 2, QChar('0'));
    }
}

bool DominoDevice::testChannelRegisters()
{
    // Channel registers test
    bool test_ok = true;
    QMap<quint16, quint16> m;
    for(int c=0; c<2; c++)
        for(int i=0; i<8; i++) {
            if (i == MEM_CH_WR_ADDR) continue; // read-only register
            if (i == MEM_CH_ADC_DATA) continue; // read-only register
            if (i == MEM_CH_ADC_PATTERN_MISMATCH_CNT) continue; // read-only register
            quint16 a = c*8+i;
            quint16 d = 1+a;
            m[a] = d;
            WriteChReg(c, i, d);
//            qDebug() << QString("Write channel %1 reg %2 = %3").arg(c).arg(i).arg(d);
        }
    for(int c=0; c<2; c++)
        for(int i=0; i<8; i++) {
            quint16 a = c*8+i;
            quint32 d = ReadChReg(c, i);
            bool ok = (d == m[a]);
            if (i == MEM_CH_WR_ADDR) ok = true; // read-only register
            if (i == MEM_CH_ADC_DATA) ok = true; // read-only register
            if (i == MEM_CH_ADC_PATTERN_MISMATCH_CNT) ok = true; // read-only register
            if (!ok) test_ok = false;
            if (!ok)
                qWarning() << QString("Read channel %1 reg %2 = %3, expected %4").arg(c).arg(i).arg(d).arg(m[a]);
//            else
//                qDebug() << QString("Read channel %1 reg %2 = %3").arg(c).arg(i).arg(d);
        }
    if (test_ok)
        qDebug() << "Channel registers test Ok";
    else
        qWarning() << "Channel registers test FAILED";

    return test_ok;
}

void DominoDevice::on_setupChanged(const DominoSettings &in)
{
    setup = in;
    if (isConnected())
        writeSettings(setup);
}

void DominoDevice::writeSettings(const DominoSettings &in)
{
    quint64 serial_id = getDeviceSerialNumber();
 //   qDebug() << "serial id = " << hex << serial_id;

    setup = in;
    try {
        if (!settingsCacheValid || (settingsCache.en != setup.en) || (settingsCache.invert_thr_trig != setup.invert_thr_trig)
                || (settingsCache.invert_zs_thr != setup.invert_zs_thr)) {
            for(int i=0; i<setup.nch; i++)
               ChannelSetup(i, setup.en.value(i), (setup.invert_thr_trig ? TRIG_NEGATIVE : TRIG_POSITIVE),
                           (setup.invert_zs_thr ? TRIG_NEGATIVE : TRIG_POSITIVE));
        }

        if (!settingsCacheValid || settingsCache.thr != setup.thr) {
            for(int i=0; i<setup.nch; i++)
                ChnSetThreshold(i, setup.thr.value(i));
        }

        if (!settingsCacheValid || settingsCache.trigsetup.all != setup.trigsetup.all)
            WriteTrigSetup(setup.trigsetup);

        if (!settingsCacheValid || settingsCache.trigdelay != setup.trigdelay)
            SetTrigDelay(setup.trigdelay);

        if (getDeviceId() == QtMregDevice::DEVICE_ID_ADC64) {
            if(!settingsCacheValid || settingsCache.amp_board_set != setup.amp_board_set)
                SetAmpBoard(setup.amp_board_set);
        }

        if ((getDeviceId() == QtMregDevice::DEVICE_ID_ADC64) && (setup.amp_board_set == AMP_BOARD_PCA)) {
            if (!settingsCacheValid || settingsCache.reg_pca.all != setup.reg_pca.all)
                WritePCAReg(setup.reg_pca);
        }

        if ((getDeviceId() == QtMregDevice::DEVICE_ID_ADC64) && (setup.amp_board_set == AMP_BOARD_AD5622)){
            if (!settingsCacheValid || (settingsCache.ad5622op.data_baseline != setup.ad5622op.data_baseline)
                    || (settingsCache.ad5622op.data_level != setup.ad5622op.data_level))
                WriteAD5622Command(setup.ad5622op);
        }

        if (!settingsCacheValid || settingsCache.zs_thr != setup.zs_thr ||
                settingsCache.baseline != setup.baseline) {
            for(int i=0; i<setup.nch; i++)
                ChnSetZsThreshold(i, setup.baseline.value(serial_id).value(i) + setup.zs_thr);
        }
        // DEBUG
//        for(int c=0; c<setup.nch; c++) {
//            qDebug() << QString("raw adc data [%1] = %2h")
//                        .arg(c, 2)
//                        .arg(ReadChReg(c, MEM_CH_ADC_DATA), 4, 16, QChar('0'));
//        }

        settingsCache = setup;
        settingsCacheValid = true;
    }
    catch (std::runtime_error &e) {
        settingsCacheValid = false;
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

bool DominoDevice::isMaster()
{
    return ((regRead(ADCM16_STATUS_REG)&0x7) || (regRead(REG_GTPDUAL) == 0));
}

void DominoDevice::WritePCAReg(const reg_pca_struct &reg_pca)
{  
    regWrite(REG_PCA_12, reg_pca.all);
}

void DominoDevice::WriteAD5622Command(const ad5622operation &ad5622op)
{
    regWrite(REG_AD5622_BASELINE, ad5622op.data_baseline);
    regWrite(REG_AD5622_LEVEL, ad5622op.data_level);
    regWrite(REG_AD5622_COMMAND, AD5622_WRITE_COMMAND);
}

void DominoDevice::SetAmpBoard(const quint16 board_set)
{
    regWrite(REG_AMP_SET, board_set);
}

void DominoDevice::SetTrigDelay(int delay)
{
    //    qDebug() << "trigdelay=" << setup.trigdelay;
    regWrite(REG_DEVICE_RLAT, delay-5);  //Set number of points saved after incoming trigger
}

void DominoDevice::on_timerPoll()
{
    try {
        if (isConnected()) {
            if (QDateTimeMsescTo(lastStatusTimestamp, QDateTime::currentDateTime()) > 1000) {
                readTemp();
                lastStatusTimestamp = QDateTime::currentDateTime();
            }
            if (setup.run) {
                if (!saveRunState) {
                    //            qDebug() << "Start";
                    StartRun();
                } else {
                    // read status
                    bool running = isRunning();
                    //            qDebug() << running;
                    if (!running) {
                        //                qDebug() << "Readout";
                        ReadAllChMem();
                        //                qDebug() << "Restart";
                        RestartRun();
                    }
                }
            } else {
                if (saveRunState) {
                    StopRun();
                    saveRunState = false;
                }
            }
            if (saveRunState && !setup.run)  //setup.run can change during long ReadAllChMem()
            {
                StopRun();
            }
            saveRunState = setup.run;
        } else {
            saveRunState = false;
        }
    }
    catch (std::runtime_error &e) {
        saveRunState = false;
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
    pollTimer->start(saveRunState ? 1 : pollTimeMs);
}


void DominoDevice::RestartRun()
{
    RegOpVector v;
    {
        RegOp op;
        op.addr = REG_DEVICE_CTRL;
        op.data = 0; // clear RUN
        op.write = true;
        v.push_back(op);
    }
    {
        RegOp op;
        op.addr = REG_DEVICE_CTRL;
        op.data = REG_DEVICE_CTRL_BIT_RUN; // set RUN
        op.write = true;
        v.push_back(op);
    }
    regOpExec(v);
}

void DominoDevice::StartRun()
{
    regWrite(REG_DEVICE_CTRL, 0);   // clear RUN
    regWrite(REG_DEVICE_CTRL, REG_DEVICE_CTRL_BIT_RUN);     // set RUN
}

void DominoDevice::StopRun()
{
    regWrite(REG_DEVICE_CTRL, 0);   // clear RUN
}

void DominoDevice::ReadChMem(int ch, const quint32 addr_offset)
{
    if (!isConnected())
        return;
    // two adc words per memory address
    const int ramSize = getChannelMaxSampleCount() / 2;
    const quint32 baseAddr = chBaseMemAddr(ch);
    quint32 ch_reg_stop_addr = (ReadChReg(ch, MEM_CH_WR_ADDR) - addr_offset);
    int wordsToRead = (setup.readcellnumber / 2) + ((setup.readcellnumber|ch_reg_stop_addr) % 2);
    if (wordsToRead > ramSize) wordsToRead = ramSize;
    int stopAddr = ((ch_reg_stop_addr / 2) + (ch_reg_stop_addr % 2)) % ramSize;
    int startAddr = stopAddr - wordsToRead;
    if (startAddr < 0) startAddr += ramSize;
//    qDebug() << "Start address " << startAddr;
//    qDebug() << "Stop  address " << stopAddr;

    QVector<quint32> fullrd;
    fullrd.reserve(wordsToRead); // reserve memory, vector size = 0
    int numRead = 0;
    while (numRead < wordsToRead) {
        int start = (startAddr + numRead) % ramSize;
        int stop = (startAddr + wordsToRead);
        if (start > stop) stop = ramSize;
        int count = stop-start;
//        qDebug() << QString("%1: Read %2 - %3").arg(numRead).arg(start).arg(stop);
        const QVector<quint32> &rbuf = QVector<quint32>::fromStdVector(memReadBlk(baseAddr | start, count));
        if (count != rbuf.size())
        {
            qCritical() << "read memory size mismatch";
            return;
        }
        // append
        fullrd += rbuf;
        numRead += rbuf.size();
    }

    dominoData.rawData[ch].clear();
    for (size_t i=0; i<fullrd.size(); i++)
    {
        qint16 veven = (qint16)((fullrd[i] >> 16) & 0xFFFF) - 0x8000;
        qint16 vodd  = (qint16)((fullrd[i] >> 0)  & 0xFFFF) - 0x8000;
        dominoData.rawData[ch].push_back(veven);
        dominoData.rawData[ch].push_back(vodd);
    }

//    qDebug() << "Ch " << ch << ": " << hex << fullrd;

    if (ch_reg_stop_addr % 2)  //removing first-last data for even-odd start addr. and number of readcell
    {
        dominoData.rawData[ch].pop_back();
        if (!(setup.readcellnumber % 2)) dominoData.rawData[ch].pop_front();
    }
    else if (setup.readcellnumber % 2) dominoData.rawData[ch].pop_front();
}


void DominoDevice::ReadAllChMem()
{
    dominoData.rawData.clear();
    quint32 master_slave_addr_offset = 0;  //addr. offset in 16bit data(adc samples), 0 for master
    if (!isMaster()) master_slave_addr_offset = MASTER_SLAVE_TRIG_DELAY;
    quint64 zs_events = regRead64(REG_ZS_EVENTS);

    for (int i=0; i<getNumberOfChannels(); i++) {
        if (setup.en[i] && (((zs_events >> i)&0x1) || !setup.zs_en)) {
            ReadChMem(i, master_slave_addr_offset);
        } else {
            dominoData.rawData[i]; // empty entry must be present for correct channel numbering
        }
    }
    dominoData.serial_id = getDeviceSerialNumber();
    dominoData.run_ts = readTimestampRun();
    dominoData.trig_ts = readTimestampTrig();
//    quint64 current_ts = readTimestampCurrent();
//    qDebug() << QString("[%1] TS: current %2, run %3, trig %4")
//                .arg(dominoData.serial_id, 4, 16, QChar('0'))
//                .arg(current_ts, 16, 16, QChar('0'))
//                .arg(dominoData.run_ts, 16, 16, QChar('0'))
//                .arg(dominoData.trig_ts, 16, 16, QChar('0'));
    dominoData.EventNumber++;
//    dominoData.timestamp = QDateTime::currentDateTime();
    Q_EMIT dataAcquired(dominoData);
}


quint16 DominoDevice::ReadChReg(int ch, quint16 addr)
{
    Q_ASSERT(addr < BOARD_REG_COUNT);
    quint32 read_addr = MEM_BIT_SELECT_CTRL;   // bit13==1 (bus 15:0) - register operation
    read_addr |= addr;
    read_addr |= chBaseMemAddr(ch);
    quint32 value = memRead(read_addr);
    //    qDebug() << QString("ReadChReg: memRead(%1, %2)").arg(read_addr, 8, 16, QChar('0')).arg(value, 8, 16, QChar('0'));
    return value;
}

void DominoDevice::WriteChReg(int ch, quint16 addr, quint16 data)
{
    Q_ASSERT(addr < BOARD_REG_COUNT);
    quint32 write_addr = MEM_BIT_SELECT_CTRL;   // bit13==1 (bus 15:0) - register operation
    write_addr |= addr;
    write_addr |= chBaseMemAddr(ch);
    //    qDebug() << QString("WriteChReg: memWrite(%1, %2)").arg(write_addr, 8, 16, QChar('0')).arg(quint32(data), 8, 16, QChar('0'));
    memWrite(write_addr, quint32(data));
}

bool DominoDevice::isRunning()
{
    return 0 != (regRead(REG_RUN_STATUS) & REG_RUN_STATUS_BIT_RUNNING);
}

void  DominoDevice::DacMax550Write(quint8 addr, quint8 data)
{
    quint16 wd = (addr<<8) | data ;
    {
        regWrite(REG_DAC_MAX550_1, wd);
        regWrite(REG_DAC_MAX550_2, wd);
    }
}

void DominoDevice::ChannelSetup(int ch, bool thr_trig_en, TrigPolarity thr_trig_polarity, TrigPolarity zs_thr_polarity)
{
    quint16 ch_ctrl = 0;
    if (thr_trig_en)
        ch_ctrl |= MEM_CH_CTRL_BIT_EN;
    if (thr_trig_polarity == TRIG_NEGATIVE)
        ch_ctrl |= MEM_CH_CTRL_BIT_THR_TRIG_INVERT;
    if (zs_thr_polarity == TRIG_NEGATIVE)
        ch_ctrl |= MEM_CH_CTRL_BIT_ZS_THR_INVERT;
    WriteChReg(ch, MEM_CH_CTRL, ch_ctrl);
}

void DominoDevice::ChnSetZsThreshold(int ch, int thr)
{
    thr += 0x8000;
    if (thr < 0) thr = 0;
    if (thr > 0xFFFF) thr = 0xFFFF;
    WriteChReg(ch, MEM_CH_ZS_THR, thr);
}

void DominoDevice::ChnSetThreshold(int ch, int thr)
{
    thr += 0x8000;
    if (thr < 0) thr = 0;
    if (thr > 0xFFFF) thr = 0xFFFF;
    WriteChReg(ch, MEM_CH_THR, thr);
}

void DominoDevice::WriteTrigSetup(reg_trig_ctrl trigsetup)
{
    reg_trig_ctrl slavetrigsetup;
    slavetrigsetup.all = 0;
    slavetrigsetup.bit.nim_en = 1;
    regWrite(REG_TRIG_CTRL, trigsetup.all);
}

bool DominoDevice::AdcSpiWrite(quint8 reg, quint8 data, bool readback)
{
    quint16 senddata = 0;
    senddata = (reg<<8) | data;
    regWrite(REG_ADC_SPI, senddata);

    if (!readback) return true; //Write complete, start Read if needed

    int errnum = 0;
    bool testok  = true;
    QVector<quint16> spi_rd;
    const int MAX_ERR_NUM = 16;

    do
    {
        senddata = ((reg+1)<<8);  //write even register, read odd register (read=write+1 reg number)
        regWrite(REG_ADC_SPI, senddata);
        spi_rd.clear();
        spi_rd.push_back(regRead(REG_ADC12_SPI_READ));
        spi_rd.push_back(regRead(REG_ADC34_SPI_READ));
        spi_rd.push_back(regRead(REG_ADC56_SPI_READ));
        spi_rd.push_back(regRead(REG_ADC78_SPI_READ));
        testok = true;
        for (int i=0; i<spi_rd.size(); i++)
        {
            if ((data!=(spi_rd[i]&0xFF)) || (data!=((spi_rd[i]>>8)&0xFF)))
            {
//                qDebug() << "SPI R/W FAIL cycle " << errnum << "i=" << i << "data=" << hex << data << " spi_rd[i]=" << hex << spi_rd[i];
                errnum++;
                testok = false;
            }
            else
            {
//                qDebug() << "SPI R/W OK, i=" << i << " data = " << hex << data;
                testok &= true;
            }
        }
    } while (!testok && errnum < MAX_ERR_NUM);
    if (errnum >= MAX_ERR_NUM) qDebug() << "More ERRORS!!!";
        else qDebug() << "SPI R/W OK";
    return testok;
}

QVector<quint16> DominoDevice::AdcSpiRead(quint8 reg)
{
    QVector<quint16> spi_rd;
    quint16 senddata = 0;
    senddata = ((reg+1)<<8);  //write even register, read odd register (read=write+1 reg number)
    regWrite(REG_ADC_SPI, senddata);
    spi_rd.clear();
    spi_rd.push_back(regRead(REG_ADC12_SPI_READ));
    spi_rd.push_back(regRead(REG_ADC34_SPI_READ));
    spi_rd.push_back(regRead(REG_ADC56_SPI_READ));
    spi_rd.push_back(regRead(REG_ADC78_SPI_READ));
    qDebug() << "SPI Read " << hex << spi_rd;
    return spi_rd;
}

void DominoDevice::SetAdcCustomPattern(quint16 data)
{
//    qDebug() << "Setting pattern" << hex << data;
    data &= ADC_MASK;
    AdcSpiWrite(ADC_SPI_PATTERN_LOW_REG, (data&0xFF), false);
    AdcSpiWrite(ADC_SPI_PATTERN_HIGH_REG, (data>>8), false);
}

void DominoDevice::SetAdcCustomPatternMode(bool mode)
{
    if (mode) AdcSpiWrite(ADC_SPI_DECIMATOR_CTRL_REG, 0x3, true);
    else  AdcSpiWrite(ADC_SPI_DECIMATOR_CTRL_REG, 0, true);
}

void DominoDevice::SetAdcClkDelay(const QVector<int> delay)
{
    quint8 adc_mask = ADC_CLK_CTRL_ALL_ADCS;
    int counter = 0;
    ResetAdcClkDelay();
    do
    {
        for(int i=0; i<delay.size(); i++)
            if(delay[i] == counter)
                adc_mask &= ~(1<<i);
//        qDebug() << "counter = " << counter << "adc mask =" << hex << adc_mask;
        ChangeAdcClkDelay(true, adc_mask, 1);
        counter++;
    }
    while((adc_mask != 0) && (counter < 65));   //max tap number == 64
}

bool DominoDevice::FindAdcClkDelay()
{
    QVector< QVector<bool> > ptr; //pattern test result
    QVector<int> adc_delay_value;

    if (getDeviceId() != DEVICE_ID_ADC64) return true;  //calibrate only for ADC64 board

    qDebug() << "Starting search...";
    ResetAdc();
    ResetAdcClkDelay();
    SetAdcCustomPatternMode(true);
    for(int i=0; i<42; i++)  //scanning for 1 full bit clk time (~3.3ns)
    {
        QVector<bool> test = AdcPatternTest();
        if(ptr.size()!=test.size()) ptr.resize(test.size());
        for(int j=0; j<ptr.size(); j++) ptr[j].push_back(test[j]);
        ChangeAdcClkDelay(true, ADC_CLK_CTRL_ALL_ADCS, 1);  //increase by 1 all 8 adcs
    }
    SetAdcCustomPatternMode(false);
    for(int i=0; i<ptr.size(); i++)
        qDebug() << "Test result vector " << i << " = " << ptr[i];

    for (int i=0; i<ptr.size(); i++)
        adc_delay_value.push_back(FindTruePlateauCenter(ptr[i]));
    qDebug() << "adc_delay_value vector" << adc_delay_value;
    for (int i=0; i<adc_delay_value.size(); i++)
        if (adc_delay_value[i] == 0) return false;  //if at least one center of true plateau was not found, return false
    SetAdcClkDelay(adc_delay_value);
    return true;
}

int DominoDevice::FindTruePlateauCenter(const QVector<bool> v)
{
    int start_search_index = 0;
    if (v.size()<2) return 0;
    for(int i=2; i<v.size(); i++)  //loking for 2 false, then start search foe true plateau
        if (!v[i] && !v[i-1])
        {
            start_search_index = i;
            break;
        }
    if (start_search_index == 0) return 0;  //if it was unable to detect 2 seq. false, then search failed
    int true_plateau_start = 0;
    int true_plateau_center = 0;
    bool true_plateau_start_found = false;
    bool true_plateau_end_found = false;
    for(int i=start_search_index; i<v.size(); i++)
    {
        if(v[i] && v[i-1] && v[i-2] && !true_plateau_start_found)  //looking for 3 seq. true - start of true plateau
        {
            true_plateau_start_found = true;
            true_plateau_start = i-2;
        }
        if (true_plateau_start_found && !v[i])  //looking for first false after at least 3 true
        {
            true_plateau_end_found = true;
            true_plateau_center = (i-1+true_plateau_start) / 2;
            break;
        }
    }
    if (true_plateau_start_found && true_plateau_end_found) return true_plateau_center;
    else return 0;
}

QVector<bool> DominoDevice::AdcPatternTest()
{
    quint16 spi_test_pattern = 0;
    quint16 adc_ch_test_pattern = 0;
    QVector<bool> testok;

    for (int n=0; n<2; n++)  //testing 2 random pattern on each clock tap
    {
        spi_test_pattern = qrand() & 0xFFF;
        SetAdcCustomPattern(spi_test_pattern);
        adc_ch_test_pattern = ((spi_test_pattern ^ ADC_HALF_VALUE) << (16 - ADC_BITS));  //arrange to 16bit data
        QVector<bool> test = TestAdcCustomPattern(adc_ch_test_pattern);
        if(testok.size()!=test.size())
        {
            testok.resize(test.size());
            for(int i=0; i<testok.size(); i++) testok[i]=true;
        }
        else for(int i=0; i<testok.size(); i++) testok[i]&=test[i];
    }
    return testok;
}

QVector<bool> DominoDevice::TestAdcCustomPattern(const quint16 pattern)
{
    const int octal_adc_channels = 8;
    const int pattern_cnt_reread_number = 1;  //reading pattern missmatch counters 2 times
    QVector<bool> test_ok;
    QVector <quint16> read_pattern_cnt;

    for (int i=0; i<(getNumberOfChannels()/octal_adc_channels); i++)
        for(int k=0; k<octal_adc_channels; k++)
        {
            read_pattern_cnt.clear();
            if(test_ok.size()<i+1) test_ok.push_back(true);
            WriteChReg(i*octal_adc_channels + k, MEM_CH_ADC_PATTERN, pattern);
            for (int j=0; j<pattern_cnt_reread_number+1; j++)
                read_pattern_cnt.push_back(ReadChReg(i*octal_adc_channels + k, MEM_CH_ADC_PATTERN_MISMATCH_CNT));
            for (int j=0; j<read_pattern_cnt.size(); j++)
            {
                if (read_pattern_cnt[j]!=read_pattern_cnt[0])
                {
                    //qDebug() << "i=" << i << hex << "pattern vector" << read_pattern_cnt;
                    test_ok[i] = false;
                    break;
                }
            }
            if (!test_ok[i]) break;
        }
    return test_ok;
}

void DominoDevice::ChangeAdcClkDelay(bool increase, quint8 channel_mask, quint16 taps)
{
    quint16 send_value = 0;
    for (int i=0; i<taps; i++)
    {
        send_value = 0;
        send_value |= increase ? ADC_CLK_CTRL_BIT_INC : 0;  //increase/decrease delay
        send_value |= channel_mask;       //change delay on selected adcs
        regWrite(REG_ADC_CLK_CTRL, send_value);
        regWrite(REG_ADC_CLK_CTRL, 0);  //work on edge
    }
}

void DominoDevice::ResetAdcClkDelay()
{
    quint16 send_value = 0;
    send_value |= ADC_CLK_CTRL_BIT_RESET;
    regWrite(REG_ADC_CLK_CTRL, send_value);
    regWrite(REG_ADC_CLK_CTRL, 0);  //work on edge
}

void DominoDevice::ResetAdc()
{
    quint16 send_value = 0;
    send_value |= ADC_CLK_CTRL_BIT_ADC_RESET;
    regWrite(REG_ADC_CLK_CTRL, send_value);
    regWrite(REG_ADC_CLK_CTRL, 0);  //work on edge
}

int DominoDevice::RegWRTest(int NCycle)
{
    int randnum =0, readnum=0;
    qDebug() << "Testing...";
//    regWrite(REG_AMP_SET, 0x1); //enable ad5622 control

    ad5622operation ad5622op;
    ad5622op.data_baseline = 0x2E9;
    ad5622op.data_level = 0x64;
    ad5622op.data_command = 0x1E0;
    WriteAD5622Command(ad5622op);

//    regWrite(REG_AD5622_COMMAND + 1,0x1234);
//    qDebug() << "regRead_l =" << hex << regRead(REG_AD5622_COMMAND);
//    qDebug() << "regRead_h =" << hex << regRead(REG_AD5622_COMMAND + 1);

/*
    for (int i=0; i<NCycle; i++)
    {
        //        randnum = (qrand() & 0xFFFF);
        randnum = i;
        //        regWrite(REG_23,randnum);
        //        readnum = regRead(REG_23);
        if (randnum!=readnum)
        {
            qDebug() << "error in r/w reg test";
            qDebug() << "cycle number is" << i << " (" << hex << i << ")";
            return(0);
        }
    }
    return 1;
*/
}
