#include "mainwindow.h"

#include "ui_mainwindow.h"
#include "dominodevice.h"
#include "DominoDeviceManager.h"
#include "dominosettings.h"
#include "EventData.h"
#include "EventWriter.h"
#include <regio/regiomlink.h>

#include <QtDebug>
#include <QTimer>
#include <QHostAddress>
#include <qxw/WaveformPlot.h>
#include <stdexcept>
#include <QFileDialog>
#include <QMessageBox>
#include <QSettings>
#include <QStatusBar>
#include <QDesktopServices>
#include <QUrl>
#include <QMap>
#include <QThread>
#include <QStringList>
#include "util/QDateTimeCompat.h"


#include "waitcursor.h"

#ifndef VERSION_STR
#define VERSION_STR "0.2"
#endif

const int MAX_HOSTNAMES = 4;

Q_DECLARE_METATYPE(QHostAddress)

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    statusLabel(new QLabel(this)),
    evLabel(new QLabel(this)),
    tempLabel(new QLabel(this)),
    idLabel (new QLabel(this)),
    dominoDevice(new DominoDeviceManager(0)),
    lastEventData(new EventData),

//    dominoDevice(new DominoDevice(0)),
    eventWriter (new EventWriter),
    deviceThread(new QThread),
    writerThread(new QThread)
{
    ui->setupUi(this);
    ui->tabWidget->setCurrentWidget(ui->tabGraph);
    statusBar()->addPermanentWidget(statusLabel, 10);
    statusBar()->addPermanentWidget(evLabel, 2);
    statusBar()->addPermanentWidget(tempLabel, 1);
    statusBar()->addPermanentWidget(idLabel, 1);
    ui->label_adccalib_stat->hide();
    qRegisterMetaType<DominoSettings>();
    qRegisterMetaType<DominoData>();
    qRegisterMetaType<QHostAddress>();
//    connect(this, SIGNAL(setupChanged(DominoSettings)), dominoDevice, SLOT(on_setupChanged(DominoSettings)));
    connect(this, SIGNAL(setupChanged(DominoSettings)), dominoDevice, SIGNAL(setupChanged(DominoSettings)));
    connect(dominoDevice, SIGNAL(deviceConnected()), SLOT(onDeviceConnect()));
    connect(dominoDevice, SIGNAL(deviceDisconnected()), SLOT(onDeviceDisconnect()));
    connect(dominoDevice, SIGNAL(setStatus(const QString &)), SLOT(setStatus(const QString &)));
    connect(dominoDevice, SIGNAL(tempChanged(const QString &)), SLOT(setTemp(const QString &)));
    connect(dominoDevice, SIGNAL(numberOfChannelChanged(int)), SLOT(setNumberOfChannels(int)));
    connect(dominoDevice, SIGNAL(adcCalibrationStatusChanged(bool)), SLOT(setAdcCalibrationStatus(bool)));
    connect(dominoDevice, SIGNAL(adcCalibrationStart()), SLOT(setAdcCalibrationStart()));
    connect(dominoDevice, SIGNAL(monitorEventChanged(EventData)), SLOT(DrawScope(EventData)));

    QStringList edge = (QStringList() << "Rising" << "Falling");
    ui->comboBoxThrTrig->addItems(edge);
    ui->comboBoxZSThr->addItems(edge);

    readSettings();
    restoreFormValues();
    //    ui->groupBoxWRReg->hide();
    //    ui->groupBoxDAC->hide();
//    ui->pushButtonTest2->hide();
//    ui->spinBoxT1->hide(); ui->spinBoxT2->hide();
//    ui->spinBoxT3->hide(); ui->spinBoxT4->hide();
//    ui->spinBoxT5->hide(); ui->spinBoxT6->hide();
//    ui->spinBoxT7->hide(); ui->spinBoxT8->hide();
    ui->radioButtonAOffMod->hide();
    enableRunControlButtons();
    ui->checkBoxSnapshot->hide();
    ui->groupBoxMode->hide();
    ui->mainToolBar->hide();

    deviceThread->start();
    dominoDevice->moveToThread(deviceThread);

    writerThread->start();
    eventWriter->moveToThread(writerThread);

    QTimer::singleShot(10, this, SLOT(on_pushButtonConnect_clicked()));
}

MainWindow::~MainWindow()
{
    deviceThread->quit();
    writerThread->quit();
    qDebug() << "Waiting threads to shut down";
    if (!deviceThread->wait(1000))
        qWarning() << "Device thread aborted";
    if (!writerThread->wait(1000))
        qWarning() << "Writer thread aborted";
    delete deviceThread;
    delete writerThread;
    delete dominoDevice;
    delete ui;
}

void MainWindow::setupDevice()
{
    Q_EMIT setupChanged(setup);
}

void MainWindow::setNumberOfChannels(int n)
{
//    qDebug() << QString("MainWindow::setNumberOfChannels(%1)").arg(n);
    setup.resize(n);
    //    qDebug() << "Signal setNumberOfChannels, number of boards = " << dominoDevice->GetNumberOfAdc64boards();
    QMap<int, QString> map;
    for(int i=0; i<n/8; i++)
        map[i+1] = QString("Group %1 (%2-%3)").arg(i+1).arg(i*8+1).arg((i+1)*8);
    ui->comboBoxOctalADC->clear();
    int index = 0;
    ui->comboBoxOctalADC->insertItem(index++, QString("All channels (%1-%2)").arg(n>0?1:0).arg(n), 0);
    Q_FOREACH(int i, map.keys())
        ui->comboBoxOctalADC->insertItem(index++, map[i], i);

    int channelMaxSampleCount = dominoDevice->getChannelMaxSampleCount();
//    ui->spinBoxWriteCNT->setMaximum(channelMaxSampleCount);
    ui->spinBoxReadCNT->setMaximum(channelMaxSampleCount);
}

void MainWindow::setAdcCalibrationStatus(bool status)
{
    ui->label_adccalib_stat->setText(status ? "ADC CLK delay calibration OK" : "ADC CLK delay calibration FAILED");
    ui->label_adccalib_stat->setHidden(status);
}

void MainWindow::setAdcCalibrationStart()
{
    ui->label_adccalib_stat->setText("Calibrating ADC CLK Delay ...");
    ui->label_adccalib_stat->setHidden(false);
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    writeSettings();
    QMainWindow::closeEvent(event);
}

void MainWindow::on_actionQuit_activated()
{
    close();
}

void MainWindow::on_actionDocs_activated()
{
    QDesktopServices::openUrl(QUrl("http://afi.jinr.ru/ADC64"));
}

void MainWindow::on_actionCheckForUpdates_activated()
{
    QDesktopServices::openUrl(QUrl("http://afi.jinr.ru/Downloads/ADC64"));
}

void MainWindow::on_actionAbout_activated()
{
    QString rev;
#ifdef SVN_REV
    rev = QString("%1").arg(SVN_REV);
    rev = rev.remove(QRegExp("^.*\\:"));
#endif
    if (!rev.isEmpty())
        rev = QString(", rev. %1").arg(rev);
    QString s = "<p>ADC64<br>Version "+QString(VERSION_STR)+rev+"<br><br>&copy; 2011 <a href=\"http://afi.jinr.ru/\">AFI Electronics</a><br>JINR, Dubna";
    QMessageBox::about (this, "About", s);
}

void MainWindow::on_actionAboutQt_activated()
{
    qApp->aboutQt();
}

void MainWindow::setStatus(const QString &str)
{
    statusLabel->setText(str);
}

void MainWindow::onDeviceConnect()
{
    setupDevice();
    idLabel->setText(dominoDevice->getDeviceIdStr());
    ui->spinBoxAfterTrig->setMaximum(dominoDevice->getChannelMaxSampleCount());
    enableRunControlButtons();
}

void MainWindow::onDeviceDisconnect()
{
    //    QMessageBox::critical(0, "Error", "Device disconnected");
    idLabel->clear();
    tempLabel->clear();
    enableRunControlButtons();
}

void MainWindow::setTemp(const QString &str)
{
    tempLabel->setText(str);
}

//void MainWindow::DoStopTimer()
//{
//    //    qDebug() << "stop timer";
////    LocalWaitCursor wc(this);
//    dominoDevice->StopRun();
//    //    dominoDevice->ReadAllChMem();
//    //    DrawScope();
//    ui->pushButtonStart->setEnabled(true);
//    ui->pushButtonStop->setEnabled(false);
//    ui->checkBoxW2D->setEnabled(true);
//}

//void MainWindow::DoTimer()
//{
//    try
//    {
//        bool isRunning = dominoDevice->isRunning();
//        if (!isRunning)
//        {
//            dominoDevice->ReadAllChMem();
//            if (ui->checkBoxSnapshot->isChecked()) {
//                on_pushButtonStop_clicked();
//                ui->pushButtonStop->setEnabled(false);
//            } else {
//                dominoDevice->RestartRun();
//                ui->pushButtonStop->setEnabled(true);
//            }
//        }
//    }
//    catch(std::runtime_error& e)
//    {
//        qDebug() << e.what();
//    }
//}

//void MainWindow::SaveData(const DominoData &dominoData)
//{
//    if(ui->checkBoxW2D->isChecked()) {
//        const QDateTime &current = QDateTime::currentDateTime();

//        if (QDateTimeMsescTo(dominoData.timestamp, current) > 1000) { // data too old
//            qDebug() << QString("SaveData: event %1 dropped (old)").arg(dominoData.EventNumber);
//            return;
//        }
//        QMetaObject::invokeMethod(eventWriter, "SaveData", Q_ARG(DominoData, dominoData));
//        eventWriter.SaveData(dominoData);
//    }
//}

void MainWindow::DrawScope(const EventData &eventData)
{
    const QDateTime &current = QDateTime::currentDateTime();

    if (QDateTimeMsescTo(scopeDisplayTime, current) < 20) // update too fast
        return;

    if (QDateTimeMsescTo(eventData.timestamp, current) > 100) { // data too old
        qDebug() << QString("DrawScope: event %1 dropped (old)").arg(eventData.getEventNumberStr());
        return;
    }

    *lastEventData = eventData;

    WaveformDataAll mvv;
    int channelOffset = 0;
    Q_FOREACH(const DominoData &dominoData, eventData) {
        //    const DominoData &dominoData = dominoDevice->getDominoData();
        const DominoRawData &dd = dominoData.rawData;
        int octal_adc_num = ui->comboBoxOctalADC->currentIndex() - 1;
        bool displayAll = (octal_adc_num == -1);
        QMap<int, bool> displayEn;

        for(int i=octal_adc_num*8; i < octal_adc_num*8+8; i++)
        {
            if (!setup.en.value(i%8))
                continue;
            displayEn[i] = true;
        }
        Q_FOREACH(int i, dd.keys()) {
            if (!displayAll && !displayEn[channelOffset+i])
                continue;
            if (dd[i].empty())
                continue;
            WaveformData wd;
            wd.c = channelOffset+i;
            wd.name = QString("ch%1").arg(channelOffset+i+1).toStdString();

            wd.p.resize(dd[i].size());
            for(size_t j=0; j < dd[i].size(); j++)
            {
                wd.p[j].y = dd[i][j];
                wd.p[j].x = j;
            }
            mvv.push_back(wd);
        }
        channelOffset += dd.size();
//        qDebug() << "channelOffset=" << channelOffset;
    }
    ui->xPlot->setAllData(mvv);
//    evLabel->setText(QString("Ev %1").arg(eventData.getEventNumberStr()));
    evLabel->setText(QString("Ev %1").arg(eventData.FullEventNumber));
    //    dominoData.PrintRMS();
    scopeDisplayTime = QDateTime::currentDateTime();
}

void MainWindow::enableRunControlButtons()
{
    ui->pushButtonStart->setEnabled(!setup.run && dominoDevice->isConnected());
    ui->pushButtonStop->setEnabled(setup.run && dominoDevice->isConnected());
}

void MainWindow::on_pushButtonStart_clicked()
{
    if (setup.run)
        return;
    setup.run = true;
    QMetaObject::invokeMethod(dominoDevice, "clearEventNumber");
    QMetaObject::invokeMethod(dominoDevice, "clearFullEventNumber");
    setupDevice();
    enableRunControlButtons();
}

void MainWindow::on_pushButtonStop_clicked()
{
    if (!setup.run)
        return;
    setup.run = false;
    setupDevice();
    enableRunControlButtons();
}


void MainWindow::on_checkBoxW2D_toggled(bool checked)
{
    if (checked)
    {
        QString fileName = QFileDialog::getSaveFileName(this,
                                                        tr("Save File"),
                                                        eventWriter->DataFileName,
                                                        tr("Text files *.txt;;All files *.*"));
        if (!fileName.isNull()) {
            QMetaObject::invokeMethod(eventWriter, "setFileName", Q_ARG(QString, fileName));
            connect(dominoDevice, SIGNAL(dataAcquired(EventData)), eventWriter, SLOT(SaveData(EventData)));
        }
        else ui->checkBoxW2D->setChecked(false);
    } else {
        QMetaObject::invokeMethod(eventWriter, "close");
        disconnect(dominoDevice, SIGNAL(dataAcquired(EventData)), eventWriter, 0);
    }

}

void MainWindow::restoreFormValues()
{
    ui->checkBoxE1->setChecked(setup.en[0]);
    ui->checkBoxE2->setChecked(setup.en[1]);
    ui->checkBoxE3->setChecked(setup.en[2]);
    ui->checkBoxE4->setChecked(setup.en[3]);
    ui->checkBoxE5->setChecked(setup.en[4]);
    ui->checkBoxE6->setChecked(setup.en[5]);
    ui->checkBoxE7->setChecked(setup.en[6]);
    ui->checkBoxE8->setChecked(setup.en[7]);

    int thrStep = 64;
    ui->spinBoxT1->setSingleStep(thrStep);
    ui->spinBoxT2->setSingleStep(thrStep);
    ui->spinBoxT3->setSingleStep(thrStep);
    ui->spinBoxT4->setSingleStep(thrStep);
    ui->spinBoxT5->setSingleStep(thrStep);
    ui->spinBoxT6->setSingleStep(thrStep);
    ui->spinBoxT7->setSingleStep(thrStep);
    ui->spinBoxT8->setSingleStep(thrStep);

    ui->spinBoxT1->setMinimum(-0x8000);
    ui->spinBoxT1->setMaximum(0x7FFF);
    ui->spinBoxT2->setMinimum(-0x8000);
    ui->spinBoxT2->setMaximum(0x7FFF);
    ui->spinBoxT3->setMinimum(-0x8000);
    ui->spinBoxT3->setMaximum(0x7FFF);
    ui->spinBoxT4->setMinimum(-0x8000);
    ui->spinBoxT4->setMaximum(0x7FFF);
    ui->spinBoxT5->setMinimum(-0x8000);
    ui->spinBoxT5->setMaximum(0x7FFF);
    ui->spinBoxT6->setMinimum(-0x8000);
    ui->spinBoxT6->setMaximum(0x7FFF);
    ui->spinBoxT7->setMinimum(-0x8000);
    ui->spinBoxT7->setMaximum(0x7FFF);
    ui->spinBoxT8->setMinimum(-0x8000);
    ui->spinBoxT8->setMaximum(0x7FFF);

    ui->spinBoxT1->setValue(setup.thr[0]);
    ui->spinBoxT2->setValue(setup.thr[1]);
    ui->spinBoxT3->setValue(setup.thr[2]);
    ui->spinBoxT4->setValue(setup.thr[3]);
    ui->spinBoxT5->setValue(setup.thr[4]);
    ui->spinBoxT6->setValue(setup.thr[5]);
    ui->spinBoxT7->setValue(setup.thr[6]);
    ui->spinBoxT8->setValue(setup.thr[7]);

    /*
    ui->checkBoxTr1->setChecked(setup.trig[0]);
    ui->checkBoxTr2->setChecked(setup.trig[1]);
    ui->checkBoxTr3->setChecked(setup.trig[2]);
    ui->checkBoxTr4->setChecked(setup.trig[3]);
    ui->checkBoxTr5->setChecked(setup.trig[4]);
    ui->checkBoxTr6->setChecked(setup.trig[5]);
    ui->checkBoxTr7->setChecked(setup.trig[6]);
    ui->checkBoxTr8->setChecked(setup.trig[7]);
*/
    ui->radioButtonNormMode->setChecked(setup.mode == 0);
    ui->radioButtonAOffMod->setChecked(setup.mode == 1);

//    ui->spinBoxWriteCNT->setMinimum(4);
//    ui->spinBoxWriteCNT->setMaximum(setup.writecellnumber);
//    ui->spinBoxWriteCNT->setValue(setup.writecellnumber);
    ui->spinBoxReadCNT->setMinimum(4);
    ui->spinBoxReadCNT->setMaximum(setup.readcellnumber);
    ui->spinBoxReadCNT->setValue(setup.readcellnumber);
    //    dominoDevice->SetReadCount(setup.readcellnumber);
    ui->spinBoxAfterTrig->setValue(setup.trigdelay);
//    ui->spinBoxADelay->setValue(setup.analogDelay);
//    if (dominoDevice->getDeviceId() != QtMregDevice::DEVICE_ID_ADC64) {
//        ui->groupBoxShaperSetup->hide();
////        ui->tabWidget->removeTab(ui->tabWidget->indexOf(ui->tabTest));
//    }
    if (setup.trigsetup.bit.timer_en) ui->checkBoxTrigTime->setChecked(true);
    if (setup.trigsetup.bit.nim_en) ui->checkBoxTrigNIM->setChecked(true);
    if (setup.trigsetup.bit.threshold_en) ui->checkBoxTrigThreshold->setChecked(true);

    ui->spinBoxZSThr->setMinimum(-0x8000);
    ui->spinBoxZSThr->setMaximum(0x7FFF);
    int zsthrStep = 16;
    ui->spinBoxZSThr->setSingleStep(zsthrStep);

    ui->groupBoxZS->setChecked(setup.zs_en);
    ui->spinBoxZSThr->setValue(setup.zs_thr);

    ui->comboBoxThrTrig->setCurrentIndex(setup.invert_thr_trig ? 1 : 0);
    ui->comboBoxZSThr->setCurrentIndex(setup.invert_zs_thr ? 1 : 0);

    ui->spinBoxAmpBL->setValue(setup.ad5622op.data_baseline);
    ui->spinBoxAmpLVL->setValue(setup.ad5622op.data_level);

    if (setup.amp_board_set == AMP_BOARD_PCA) ui->groupBoxShaperSetup->setChecked(true);
    if (setup.amp_board_set == AMP_BOARD_AD5622) ui->groupBoxAmp64Setup->setChecked(true);
}

void MainWindow::setChannelEnabled(int n, bool value)
{
    for(int i=0; i<setup.nch / 8; i++)
        setup.en[n+i*8] = value;
    setupDevice();
}

void MainWindow::setChannelThreshold(int n, int value)
{
    for(int i=0; i<setup.nch / 8; i++)
        setup.thr[n+i*8] = value;
    setupDevice();
}

void MainWindow::setChannelTrig(int n, bool value)
{
    for(int i=0; i<setup.nch / 8; i++)
        setup.trig[n+i*8] = value;
    setupDevice();
}

void MainWindow::on_checkBoxE1_toggled(bool checked)
{ setChannelEnabled(0, checked); }
void MainWindow::on_checkBoxE2_toggled(bool checked)
{ setChannelEnabled(1, checked); }
void MainWindow::on_checkBoxE3_toggled(bool checked)
{ setChannelEnabled(2, checked); }
void MainWindow::on_checkBoxE4_toggled(bool checked)
{ setChannelEnabled(3, checked); }
void MainWindow::on_checkBoxE5_toggled(bool checked)
{ setChannelEnabled(4, checked); }
void MainWindow::on_checkBoxE6_toggled(bool checked)
{ setChannelEnabled(5, checked); }
void MainWindow::on_checkBoxE7_toggled(bool checked)
{ setChannelEnabled(6, checked); }
void MainWindow::on_checkBoxE8_toggled(bool checked)
{ setChannelEnabled(7, checked); }

void MainWindow::on_spinBoxT1_valueChanged(int value)
{ setChannelThreshold(0, value); }
void MainWindow::on_spinBoxT2_valueChanged(int value)
{ setChannelThreshold(1, value); }
void MainWindow::on_spinBoxT3_valueChanged(int value)
{ setChannelThreshold(2, value); }
void MainWindow::on_spinBoxT4_valueChanged(int value)
{ setChannelThreshold(3, value); }
void MainWindow::on_spinBoxT5_valueChanged(int value)
{ setChannelThreshold(4, value); }
void MainWindow::on_spinBoxT6_valueChanged(int value)
{ setChannelThreshold(5, value); }
void MainWindow::on_spinBoxT7_valueChanged(int value)
{ setChannelThreshold(6, value); }
void MainWindow::on_spinBoxT8_valueChanged(int value)
{ setChannelThreshold(7, value); }

void MainWindow::on_checkBoxTr1_toggled(bool checked)
{
    setup.reg_pca.bit.shaper1_1 = checked ? 1:0;
    setup.reg_pca.bit.shaper2_1 = checked ? 1:0;
    setupDevice();
}
void MainWindow::on_checkBoxTr2_toggled(bool checked)
{
    setup.reg_pca.bit.shaper1_2 = checked ? 1:0;
    setup.reg_pca.bit.shaper2_2 = checked ? 1:0;
    setupDevice();
}
void MainWindow::on_checkBoxTr3_toggled(bool checked)
{
    setup.reg_pca.bit.shaper1_3 = checked ? 1:0;
    setup.reg_pca.bit.shaper2_3 = checked ? 1:0;
    setupDevice();
}
void MainWindow::on_checkBoxTr4_toggled(bool checked)
{
    setup.reg_pca.bit.gain1_1 = checked ? 1:0;
    setup.reg_pca.bit.gain2_1 = checked ? 1:0;
    setupDevice();
}
void MainWindow::on_checkBoxTr5_toggled(bool checked)
{
    setup.reg_pca.bit.gain1_2 = checked ? 1:0;
    setup.reg_pca.bit.gain2_2 = checked ? 1:0;
    setupDevice();
}
void MainWindow::on_checkBoxTr6_toggled(bool checked)
{
    setup.reg_pca.bit.polarity1 = checked ? 1:0;
    setup.reg_pca.bit.polarity2 = checked ? 1:0;
    setupDevice();
}
void MainWindow::on_checkBoxTr7_toggled(bool checked)
{
    setup.reg_pca.bit.preamp_en1 = checked ? 1:0;
    setup.reg_pca.bit.preamp_en2 = checked ? 1:0;
    setupDevice();
}
void MainWindow::on_checkBoxTr8_toggled(bool checked)
{
    setup.trig[7] = checked;
    setupDevice();
}

void MainWindow::on_radioButtonNormMode_toggled(bool checked)
{
    setup.mode = checked ? DominoSettings::NormalMode : DominoSettings::AmpCalMode;
    setupDevice();
}

void MainWindow::on_spinBoxReadCNT_valueChanged(int newval)
{
    setup.readcellnumber=newval;
    setupDevice();
}

void MainWindow::on_spinBoxAfterTrig_valueChanged(int newval)
{
    setup.trigdelay=newval;
    setupDevice();
}

void MainWindow::readSettings()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    for(int i=0; i<MAX_HOSTNAMES; i++)
    {
        QString attrname = "boardHostAddress" + i;
        QString s = settings.value(attrname).toString();
        if(!s.isEmpty()) ui->comboBoxHostName1->addItem(s);
    }
    //    setBoardHostName();
    for(int i=0;;i++) {
        QString key = QString("setup.en.%1").arg(i);
        if (!settings.contains(key)) break;
        setup.en[i] = settings.value(key).toBool();
    }
    for(int i=0;;i++) {
        QString key = QString("setup.thr.%1").arg(i);
        if (!settings.contains(key)) break;
        setup.thr[i] = settings.value(key).toInt();
    }
    for(int i=0;;i++) {
        QString key = QString("setup.trig.%1").arg(i);
        if (!settings.contains(key)) break;
        setup.trig[i] = settings.value(key).toBool();
    }
    setup.mode = settings.value("mode", setup.mode).value<DominoSettings::Mode>();
    setup.readcellnumber = settings.value("readcellnumber", setup.readcellnumber).toInt();
    setup.trigdelay = settings.value("trigdelay", setup.trigdelay).toInt();
    setup.trigsetup.all = settings.value("trigsetup", setup.trigsetup.all).toInt();
    setup.zs_en = settings.value("zs_en", setup.zs_en).toBool();
    setup.zs_thr = settings.value("zs_thr", setup.zs_thr).toInt();
    setup.ad5622op.data_level = settings.value("ad5622_level", setup.ad5622op.data_level).toInt();
    setup.ad5622op.data_baseline = settings.value("ad5622_baseline", setup.ad5622op.data_baseline).toInt();
    setup.amp_board_set = settings.value("amp_board_set", setup.amp_board_set).toInt();
    setup.invert_thr_trig = settings.value("invert_thr_trig", setup.invert_thr_trig).toBool();
    setup.invert_zs_thr = settings.value("invert_zs_thr", setup.invert_zs_thr).toBool();
// EVADC only
//    setup.writecellnumber = settings.value("writecellnumber", setup.writecellnumber).toInt();
    settings.endGroup();
}

void MainWindow::writeSettings()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    for(int i=0; i<MAX_HOSTNAMES; i++)
    {
        if(i >= ui->comboBoxHostName1->count()) continue;
        QString attrname = "boardHostAddress" + i;
        settings.setValue(attrname, ui->comboBoxHostName1->itemText(i));
    }
    Q_FOREACH(int i, setup.en.keys())
        settings.setValue(QString("setup.en.%1").arg(i), setup.en[i]);
    Q_FOREACH(int i, setup.thr.keys())
        settings.setValue(QString("setup.thr.%1").arg(i), setup.thr[i]);
    Q_FOREACH(int i, setup.trig.keys())
        settings.setValue(QString("setup.trig.%1").arg(i), setup.trig[i]);
    settings.setValue("mode", setup.mode);
    settings.setValue("readcellnumber", setup.readcellnumber);
    settings.setValue("trigdelay", setup.trigdelay);
    settings.setValue("trigsetup", setup.trigsetup.all);
    settings.setValue("zs_en", setup.zs_en);
    settings.setValue("zs_thr", setup.zs_thr);
    settings.setValue("ad5622_baseline", setup.ad5622op.data_baseline);
    settings.setValue("ad5622_level", setup.ad5622op.data_level);
    settings.setValue("amp_board_set", setup.amp_board_set);
    settings.setValue("invert_thr_trig", setup.invert_thr_trig);
    settings.setValue("invert_zs_thr", setup.invert_zs_thr);
// EVADC only
//    settings.setValue("writecellnumber", setup.writecellnumber);
    settings.endGroup();
}


void MainWindow::on_pushButtonConnect_clicked()
{
    setBoardHostName();
}

void MainWindow::setBoardHostName()
{
    QString s = ui->comboBoxHostName1->currentText();
    int index = ui->comboBoxHostName1->findText(s);
    if (index != -1) ui->comboBoxHostName1->removeItem(index);
    ui->comboBoxHostName1->insertItem(0, s);
    ui->comboBoxHostName1->setCurrentIndex(0);
    const QStringList &sl = ui->comboBoxHostName1->currentText().split(QRegExp("[,; ]"));
    QList<QHostAddress> hostAddressList;
    Q_FOREACH(QString s, sl) {
        const QHostAddress &hostAddress = QHostAddress(s);
        if (hostAddress.isNull()) continue;
        hostAddressList += hostAddress;
    }

//    QMetaObject::invokeMethod(dominoDevice, "setHostAddress", Q_ARG(QHostAddress, hostAddress));
//    hostAddressList += hostAddress;
    QMetaObject::invokeMethod(dominoDevice, "setHostAddressList", Q_ARG(QList<QHostAddress>, hostAddressList));
}

void MainWindow::on_comboBoxOctalADC_currentIndexChanged(int index)
{
    DrawScope(*lastEventData);
}

void MainWindow::on_comboBoxThrTrig_currentIndexChanged(const QString &arg1)
{
    setup.invert_thr_trig = (arg1 == "Rising") ? false : true;
    setupDevice();
}

void MainWindow::on_comboBoxZSThr_currentIndexChanged(const QString &arg1)
{
    setup.invert_zs_thr = (arg1 == "Rising") ? false : true;
    setupDevice();
}

void MainWindow::on_checkBoxTrigTime_toggled(bool checked)
{
    setup.trigsetup.bit.timer_en = checked ? 1:0 ;
    setupDevice();
}

void MainWindow::on_checkBoxTrigNIM_toggled(bool checked)
{
    setup.trigsetup.bit.nim_en = checked ? 1:0 ;
    setupDevice();
}

void MainWindow::on_checkBoxTrigThreshold_toggled(bool checked)
{
    setup.trigsetup.bit.threshold_en = checked ? 1:0 ;
    setupDevice();
}

void MainWindow::on_spinBoxAmpBL_valueChanged(int newval)
{
    setup.ad5622op.data_baseline = newval;
    setupDevice();
}

void MainWindow::on_spinBoxAmpLVL_valueChanged(int newval)
{
    setup.ad5622op.data_level = newval;
    setupDevice();
}

void MainWindow::on_groupBoxZS_toggled(bool checked)
{
    setup.zs_en = checked;
    setupDevice();
}

void MainWindow::on_spinBoxZSThr_valueChanged(int arg1)
{
    setup.zs_thr = arg1;
    setupDevice();
}

void MainWindow::on_groupBoxShaperSetup_toggled(bool newval)
{
    if (!newval && (setup.amp_board_set == AMP_BOARD_PCA)) setup.amp_board_set = AMP_BOARD_NONE;
    if (newval) {
        setup.amp_board_set = AMP_BOARD_PCA;
        ui->groupBoxAmp64Setup->setChecked(false);
    }
}

void MainWindow::on_groupBoxAmp64Setup_toggled(bool newval)
{
    if (!newval && (setup.amp_board_set == AMP_BOARD_AD5622)) setup.amp_board_set = AMP_BOARD_NONE;
    if (newval) {
        setup.amp_board_set = AMP_BOARD_AD5622;
        ui->groupBoxShaperSetup->setChecked(false);
    }
}

void MainWindow::on_pushButtonLoadBaseline_clicked()
{
    QFile baselinefile;
    QString baselinefilename = QFileDialog::getOpenFileName(this,
                                                    tr("Open File"),"",
                                                    tr("Calib files *.bltxt;;All files *.*"));
    bool conv_ok = false;

    if (baselinefilename.isNull()) return;
    baselinefile.setFileName(baselinefilename);
    if (!baselinefile.open(QFile::ReadOnly))
    {
        qDebug() << "Failed to open file " << baselinefilename;
        return;
    }
    QTextStream Datain(&baselinefile);
    QString line;
    line = Datain.readLine();
    quint64 serial_id = line.toULongLong(&conv_ok,16);
    if(line.isNull() || !conv_ok)
    {
        qDebug() << "no serial id found";
        baselinefile.close();
        return;
    }
    int readbaseline = 0;
    int ch = 0;
    do {
    line = Datain.readLine();
    readbaseline = line.toInt(0,10);  //in case of error, readbaseline = 0;
    setup.baseline[serial_id][ch] = readbaseline;
    ch++;
    } while (!line.isNull());
    baselinefile.close();
}
